"""
Query.

Utils that can be used to optimize db queries and avoid redundant requests.
"""
from typing import Callable, Optional, Type, TypeVar, Union

from beanie import Document, PydanticObjectId, operators
from beanie.odm.fields import ExpressionField, LinkInfo

from .document import DocT
from .link import Link

RDocT = TypeVar("RDocT", bound=Document)  # Related Model Type


async def prefetch_related(item_list: list[DocT], to_attribute: str) -> None:
    """
    Optimize fetching of a related attribute of one-to-one relation.

    Fetch related attribute efficiently in order to avoid multiple queries that could kill the db.

    Example:
    ```python
    class ProductCategory(Document):
        name: str

    class Product(Document):
        category: ProductCategory
        price: float

    product_list = await Product.find().to_list()
    prefetch_related(product_list, 'category')
    ```

    If there are 70 product and 10 categories, `fetch_related` will only
    perform 1 query instead of 70 queries to retrieve category for all products.

    ```
    for product in product_list:
        print(product.category.name)
    ```
    """

    # If item_list is empty there is no need to continue.
    if not item_list:
        return

    # Find the model of the related attributed
    link_fields = type(item_list[0]).get_link_fields()
    assert link_fields
    related_field: LinkInfo = link_fields[to_attribute]
    assert issubclass(related_field.document_class, Document)
    related_model: Type[Document] = related_field.document_class

    def get_related_id(item_: Document) -> Optional[PydanticObjectId]:
        """Return the id of the related object."""
        link: Optional[Link[Document]] = getattr(item_, to_attribute)
        if link:
            return link.ref.id
        return None

    # Fetch the related attribute and map it to the each item in the item_list
    related_item_ids = list(set(get_related_id(item) for item in item_list))
    related_item_list = await related_model.find(operators.In(related_model.id, related_item_ids)).to_list()
    for item in item_list:
        link: Optional[Link[Document]] = getattr(item, to_attribute)
        if link:
            related_item = next((rel for rel in related_item_list if rel.id == link.ref.id), None)
            setattr(link, "doc", related_item)


async def prefetch_related_children(
    item_list: list[DocT],
    to_attribute: str,
    related_model: type[RDocT],
    related_attribute: str,
    filter_func: Optional[
        Callable[
            [list[RDocT], DocT],
            Union[None, RDocT, list[RDocT]],
        ]
    ] = None,
) -> None:
    """
    Optimize fetching of a related attributes of one-to-many relation.

    Fetch related attribute efficiently in order to avoid multiple queries that could kill the db.

    Example:
    ```python
    class ProductCategory(Document):
        name: str

    class Product(Document):
        category: ProductCategory
        price: float

    category_list = await ProductCategory.find().to_list()
    prefetch_related_children(
        category_list, to_attribute='products', related_model='Product', related_attribute='category'
    )
    ```

    If there are 70 product and 10 categories, `fetch_related` will only
    perform 1 query instead of 10 queries to retrieve products for all categories.

    ```python
    for category in category_list:
        for product in category.products:
            print(product.price)
    ```
    """

    def no_filter_func(related_items: list[RDocT], item: DocT) -> list[RDocT]:
        """Do not apply any filter."""
        assert item
        return related_items

    if not filter_func:
        filter_func = no_filter_func

    item_ids = list(set(item.id for item in item_list))
    related_expression: ExpressionField = getattr(getattr(related_model, related_attribute), "id")
    related_item_list = await related_model.find(
        operators.In(related_expression, item_ids), sort="-doc_meta.created"
    ).to_list()

    for item in item_list:
        related_items = []
        for rel in related_item_list:
            rel_link: Link[RDocT] = getattr(rel, related_attribute)
            if item.id == rel_link.ref.id:
                related_items.append(rel)
        setattr(item, to_attribute, filter_func(related_items, item))


def prepare_search_string(search_text: str) -> str:
    """Clean and reformat the search string."""
    res = search_text.strip()
    if "@" in res and not '"' in res:
        res = f'"{res}"'
    return res



"""
Test Crons.

Helpers to make it easier to create test cases for crons.
"""
import typing

import beanie
from beanie.odm.queries.find import FindMany


def get_filter_queryset_dummy() -> typing.Callable[[FindMany[typing.Any]], FindMany[typing.Any]]:
    """Return initial queryset without filtering."""

    def filter_nothing(queryset: FindMany[typing.Any]) -> FindMany[typing.Any]:
        """Filter queryset."""
        return queryset

    return filter_nothing


def get_filter_queryset_for_merchant(
    model_class: type[beanie.Document], merchant_id: beanie.PydanticObjectId
) -> typing.Callable[[FindMany[typing.Any]], FindMany[typing.Any]]:
    """Filter the queryset to the test merchant.

    Use this filter on a queryset returning a list of Merchant or objects related to Merchants.
    :model_class: The class of the data being filtered.
    """

    def filter_merchant(queryset: FindMany[typing.Any]) -> FindMany[typing.Any]:
        """Filter queryset to a specific merchant."""
        return queryset.find_many(model_class.id == merchant_id)

    def filter_related_objects(queryset: FindMany[typing.Any]) -> FindMany[typing.Any]:
        """Filter queryset to object related to a specific merchant."""
        if hasattr(model_class, "merchant"):
            return queryset.find_many(model_class.merchant.id == merchant_id)
        raise NotImplementedError

    if hasattr(model_class, "merchant"):
        return filter_related_objects

    if hasattr(model_class, "beans_card_id"):
        return filter_merchant

    raise NotImplementedError



"""Patch used for overload Beanie."""
# mypy: ignore-errors

from typing import Any, List, Optional, Union, get_args, get_origin

from beanie.odm.fields import LinkInfo, LinkTypes
from beanie.odm.registry import DocsRegistry
from beanie.odm.utils.init import Initializer as Initializer  # pylint: disable=useless-import-alias
from pydantic.fields import FieldInfo

from sap.typing import UnionType

from .link import Link as MLink

original__detect_link = Initializer.detect_link


def detect_link(self, field: FieldInfo, field_name: str) -> Optional[LinkInfo]:
    """
    Detect link and returns LinkInfo if any found.

    :param field: FieldInfo
    :param field_name: str
    :return: Optional[LinkInfo]
    """
    origin = get_origin(field.annotation)
    args = get_args(field.annotation)

    def get_link_info(link_type: str, type_args: list[Any]):
        return LinkInfo(
            field_name=field_name,
            lookup_field_name=field_name,
            document_class=DocsRegistry.evaluate_fr(type_args[0]),  # type: ignore
            link_type=link_type,
        )

    if origin is MLink:
        return get_link_info(LinkTypes.DIRECT, type_args=args)

    if origin is List and get_origin(args[0]) is MLink:
        return get_link_info(LinkTypes.LIST, type_args=get_args(args[0]))

    if origin in [Union, UnionType] and get_origin(args[0]) is MLink:
        return get_link_info(LinkTypes.OPTIONAL_DIRECT, type_args=get_args(args[0]))

    return original__detect_link(self, field=field, field_name=field_name)


Initializer.detect_link = detect_link



"""
Beanie.

This package regroups all common helpers exclusively for Beanie.
Learn more about https://github.com/roman-right/beanie
"""
from .document import Document
from .link import Link
from .patch import Initializer
from .query import prefetch_related, prefetch_related_children, prepare_search_string

__all__ = [
    "Link",
    "Document",
    "prefetch_related",
    "prefetch_related_children",
    "prepare_search_string",
    # Patching
    "Initializer",
]



"""Utils for pagination."""
from typing import Any, Generic, Optional, TypedDict, TypeVar

from fastapi import Request
from pydantic import BaseModel

from . import utils

PageDataT = TypeVar("PageDataT")


class BeanieQueryParams(TypedDict):
    """Attribute of beanie query params."""

    limit: int
    skip: int
    sort: str


class CursorInfo:
    """
    Contains information on how the list should paginated.

    This is a fake cursor paginator.
    """

    offset: int = 0
    limit: int = 10
    limit_max: int = 250
    sort: str = "-doc_meta.created"

    def __init__(self, request: Request) -> None:
        """Initialize the cursor info."""
        self.limit = int(request.query_params.get("limit", self.limit))
        cursor_str = request.query_params.get("cursor", "")
        try:
            limit, offset = utils.base64_url_decode(cursor_str).split(",")
        except ValueError:
            pass
        else:
            self.limit, self.offset = int(limit), int(offset)

        self.limit = min(self.limit, self.limit_max)

    def get_beanie_query_params(self) -> BeanieQueryParams:
        """Return params to apply to the database query when using beanie."""
        return {
            "limit": self.limit,
            "skip": self.offset,
            "sort": self.sort,
        }

    def get_next(self) -> Optional[str]:
        """Get the cursor to paginate forward."""
        offset = self.offset + self.limit
        return utils.base64_url_encode(f"{self.limit},{offset}")

    def get_previous(self) -> Optional[str]:
        """Get the cursor to paginate backward."""
        offset = self.offset - self.limit
        if offset <= 0:
            return None
        return utils.base64_url_encode(f"{self.limit},{offset}")


class PaginatedData(BaseModel, Generic[PageDataT]):
    """Represent the structure of an API paginated list response."""

    object: str = "list"
    count: int
    next: Optional[str]
    previous: Optional[str]
    data: list[Any]


class PaginatedResponse(TypedDict):
    """
    Define a standard paginated response.

    PaginatedResponse has same structure as PaginatedData.
    """

    object: str
    count: int
    next: Optional[str]
    previous: Optional[str]
    data: list[dict[str, Any]]



"""
Test Utils.

Test xlib.utils functions.
"""
import typing

import pytest
from async_asgi_testclient import TestClient

from fastapi import Request

from AppMain.asgi import app
from sap.fastapi import utils

if typing.TYPE_CHECKING:
    from pydantic.error_wrappers import ErrorDict


@pytest.mark.parametrize(
    ("data_input", "data_output"),
    [
        (
            [{"loc": ("a",), "msg": "message", "type": "value_error.str.regex"}],
            {"a": {"msg": "message", "type": "value_error.str.regex"}},
        ),
        (
            [{"loc": ("a", "b"), "msg": "message", "type": "value_error.str.regex", "ctx": {"k": "v"}}],
            {"a": {"b": {"msg": "message", "type": "value_error.str.regex", "ctx": {"k": "v"}}}},
        ),
        (
            [{"loc": ("a", "b", "c"), "msg": "message", "type": "value_error.str.regex"}],
            {"a": {"b": {"c": {"msg": "message", "type": "value_error.str.regex"}}}},
        ),
    ],
)
def test_pydantic_format_errors(data_input: list["ErrorDict"], data_output: dict[str, dict[str, typing.Any]]) -> None:
    """Test that output matches func(input)."""
    result = utils.pydantic_format_errors(data_input)
    assert result == data_output


@pytest.mark.asyncio
async def test_flash_messages() -> None:
    """Assert that flash messages can be created and displayed."""
    alert_input_list = [
        ("This is test message: info", utils.FlashLevel.INFO),
        ("This is test message: error", utils.FlashLevel.ERROR),
    ]

    @app.get("/test-append-flash-message/")
    async def append_flash_messaging(request: Request) -> dict[str, str]:
        for message, level in alert_input_list:
            utils.Flash.add_message(request, message, level=level)
        return {"result": "Ok"}

    @app.get("/test-fetch-flash-message/")
    async def fetch_flash_messaging(request: Request) -> dict[str, typing.Any]:
        return {"result": utils.Flash.get_messages(request)}

    async with TestClient(app) as client:
        response = await client.get("/test-fetch-flash-message/")
        assert response.json()["result"] == []

        await client.get("/test-append-flash-message/")

        response = await client.get("/test-fetch-flash-message/")
        alert_output_list = response.json()["result"]

    for alert_input, alert_output in zip(alert_input_list, alert_output_list):
        assert alert_input[0] == alert_output["message"]
        assert alert_input[1].value == alert_output["level"]



"""
Lambdas refers to async background tasks.

They run code in response to events which are typically messages sent to a queue.
"""

# mypy: disable-error-code="import-untyped"

import asyncio
from typing import Any, ClassVar

import celery
import celery.bootsteps
import celery.worker.consumer
import kombu
from kombu.transport.base import StdChannel  # Channel

from sap.loggers import logger

from .packet import SignalPacket
from .utils import match_amqp_topics


class LambdaTask(celery.Task):  # type: ignore[misc]
    """
    A lambda task is a task that run on a specific event, usually after receiving a packet (message).

    The Lambda will be connected to an AMQP Queue and will listen
    to packets sent to that queue that matches the packet's topic pattern.
    """

    time_limit: int = 60 * 1  # 1 minutes
    packet: SignalPacket

    def __init__(self, **kwargs: Any) -> None:
        """Initialize lambda arguments."""
        self.name = self.get_name()

    def get_name(self) -> str:
        """Return a human-readable name for this lambda."""
        return self.__module__.split(".lambdas", maxsplit=1)[0] + "." + str(self.__name__)

    async def handle_process(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Perform pre-check such as authentication and run the task."""
        raise NotImplementedError

    async def test_process(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Call this method to launch the task in test cases."""
        return await self.handle_process(*args, **kwargs)

    def run(self, *args: Any, **kwargs: Any) -> Any:
        """Run the task."""
        logger.debug("Running task=%s args=%s kwargs=%s", self.get_name(), str(args), str(kwargs))
        return asyncio.run(self.handle_process(*args, **kwargs))


def register_lambda(lambda_task_class: type[LambdaTask]) -> LambdaTask:
    """Register the Lambda Task to make it discoverable by task runner (celery)."""
    return lambda_task_class()


class LambdaWorker(celery.bootsteps.ConsumerStep):  # type: ignore[misc]
    """Celery worker that consumes packets (messages) sent to lambda queues."""

    packets: list[SignalPacket] = []
    name: ClassVar[str] = ""

    def _get_queues(self, channel: StdChannel) -> list[kombu.Queue]:
        """Retrieve the list of AMQP queues associated to each packet signal."""
        queue_list: list[kombu.Queue] = []
        for packet in self.packets:
            # declare fallback queue
            params = packet.queue_get_params(task_name=self.name, is_fallback=True)
            params["exchange"] = kombu.Exchange(name=params["exchange"], type="topic", channel=channel, durable=True)
            queue_fallback = kombu.Queue(**params, channel=channel)
            queue_fallback.declare()

            # declare primary queue
            params = packet.queue_get_params(task_name=self.name, is_fallback=False)
            params["exchange"] = kombu.Exchange(name=params["exchange"], type="topic", channel=channel, durable=True)
            queue_primary = kombu.Queue(**params, channel=channel)
            queue_primary.declare()

            # only listen to primary queue
            queue_list.append(queue_primary)

        return queue_list

    def get_consumers(self, channel: StdChannel) -> list[kombu.Consumer]:
        """
        Create packet consumers.

        The consumers are the entrypoint of
        the application once celery starts receiving messages.
        """
        return [
            kombu.Consumer(
                channel,
                queues=self._get_queues(channel),
                callbacks=[self.consume],
                accept=["json"],
                prefetch_count=10,
            )
        ]

    def consume(self, body: dict[str, Any], message: kombu.Message) -> None:
        """
        Run the celery worker and consume messages.

        This is the entrypoint of the application once celery starts receiving messages.
        All packets received are sent to this function that will acknowledge reception and dispatch
        to registered Lambda tasks.
        """
        topic = message.delivery_info["routing_key"]
        headers = message.headers
        is_retry = headers and headers.get("x-death")
        if is_retry:
            logger.debug(
                "Consuming worker name=%s topic=%s body=%s headers=%s", self.name, topic, str(body), str(headers)
            )
        try:
            self._propagate_signal(body, message)
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception(exc)
            message.reject()
        else:
            message.ack()

    def _propagate_signal(self, body: dict[str, Any], message: kombu.Message) -> None:
        """
        Execute each lambda task that registered to that packet signal.

        Lambda tasks are all executed asynchronously and simultaneously through other background celery workers.
        Sometimes this can leads to duplicate key errors or integrity errors.
        """
        topic = message.delivery_info["routing_key"]
        for task in self.get_task_list():
            if match_amqp_topics(task.packet.topic, topic):
                # logger.debug(f"Matching task.packet.topic={task.packet.topic} topic={topic} task={task.get_name()}")
                identifier = body.get("identifier") or body.get("card_pid") or body.get("clover_id")
                task.apply_async(args=(identifier,), kwargs=body["kwargs"], time_limit=60)

    def get_task_list(self) -> list[LambdaTask]:
        """Retrieve the list of lambda tasks to execute."""
        raise NotImplementedError



"""
Test Beans Client.

Test Beans API client wrapper and common errors.
"""

import httpx
import pytest

from AppMain.settings import AppSettings
from sap.rest import BeansClient, rest_exceptions

integration_params = AppSettings.TOKENIFY
testcases_params = AppSettings.TESTCASES


@pytest.mark.asyncio
async def test_beans_api_get_access_token_invalid() -> None:
    """Test retrieving access_token on oauth."""
    with pytest.raises(rest_exceptions.Rest404Error):
        await BeansClient.get_access_token(
            "invalid_oauth_code",
            beans_public=integration_params.beans_public,
            beans_secret=integration_params.beans_secret,
        )


@pytest.mark.asyncio
async def test_beans_api_get() -> None:
    """Test retrieving a RuleType object on Beans API."""
    data = await BeansClient(access_token="").get("liana/rule_type/rule:liana:currency_spent")
    assert isinstance(data.response, httpx.Response)
    assert data["object"] == "rule_type"


@pytest.mark.asyncio
async def test_beans_api_post() -> None:
    """Test creating object without authentication."""
    with pytest.raises(rest_exceptions.Rest401Error):
        await BeansClient(access_token="").post("liana/rule/", json={"type": "rule:liana:currency_spent"})

    with pytest.raises(rest_exceptions.Rest400Error):
        await BeansClient(access_token=testcases_params.beans_access_token).post(
            "liana/credit/", json={"rule": "rule:liana:currency_spent"}
        )


@pytest.mark.asyncio
async def test_beans_api_put() -> None:
    """Test updating object without permission."""
    with pytest.raises(rest_exceptions.Rest403Error):
        await BeansClient(access_token=testcases_params.beans_card_id).put(
            "liana/rule/rule:liana:currency_spent", json={}
        )


@pytest.mark.asyncio
async def test_beans_api_delete() -> None:
    """Test method not allowed."""
    with pytest.raises(rest_exceptions.Rest405Error):
        await BeansClient(access_token="").delete("liana/rule_type/rule:liana:currency_spent")


@pytest.mark.asyncio
async def test_beans_api_invalid_path() -> None:
    """Test query a bad path."""
    with pytest.raises(rest_exceptions.Rest405Error):
        await BeansClient(access_token="").get("liana/rule_type_bad")



"""
# Exceptions.

Known exceptions that can occur while using the app.
"""


import typing

import httpx

# ---------------------------------------
# --------- Rest API ERROR -------------
# ---------------------------------------

# Those errors are returned by the Rest API


class RestAPIError(Exception):
    """Base Rest API error."""

    code: int = 0
    message: str = ""
    data: typing.Optional[dict[str, str]] = None
    request: httpx.Request
    response: httpx.Response

    def __repr__(self) -> str:
        """Display a string representation of the object."""
        return f"<{self.__class__.__name__}: {self.message}>"

    def __str__(self) -> str:
        """Display a string representation of the error."""
        return str(self.message)

    def __init__(
        self,
        *args: object,
        request: httpx.Request,
        response: httpx.Response,
        data: typing.Optional[dict[str, typing.Any]] = None,
    ) -> None:
        """Add error data."""
        super().__init__(*args)
        self.data = data
        self.request = request
        self.response = response

        if not data:
            return

        # verify if `error` is in data and if it is not empty
        if error_ := data.get("error"):
            if isinstance(error_, str):
                self.message = error_
            elif isinstance(error_, list) and isinstance(error_[0], str):
                self.message = ". ".join(error_)
            elif isinstance(error_, dict) and "message" in error_:
                self.message = error_["message"]
            return

        # verify if `message` is in data and if it is not empty
        if message_ := data.get("message"):
            if isinstance(message_, str):
                self.message = message_
            return

        self.message = str(data)


class Rest400Error(RestAPIError):
    """Data invalid."""

    code = 400
    message = "Invalid data"


class Rest401Error(RestAPIError):
    """Access token declined."""

    code = 401
    message = "Authentication refused."


class Rest403Error(RestAPIError):
    """Insufficient permission to access the requested data."""

    code = 403
    message = "Permission error."


class Rest404Error(RestAPIError):
    """Path or object does not exist."""

    code = 404
    message = "Data not found."


class Rest405Error(RestAPIError):
    """The method is not allowed is on this path."""

    code = 405
    message = "Method or path not allowed."


class Rest409Error(RestAPIError):
    """There was a conflict due to a duplicate transaction."""

    code = 409
    message = "Conflict or duplicate transaction."


class Rest503Error(RestAPIError):
    """The Rest server is down."""

    code = 503
    message = "The Rest server is unreachable."


RestErrorMap = {
    400: Rest400Error,
    422: Rest400Error,
    401: Rest401Error,
    403: Rest403Error,
    404: Rest404Error,
    405: Rest405Error,
    409: Rest409Error,
    # Server Error
    500: Rest503Error,
    501: Rest503Error,
    502: Rest503Error,
    503: Rest503Error,
    504: Rest503Error,
    505: Rest503Error,
}



"""
Exceptions.

Group common logic exceptions such as ValidationError
"""
from fastapi import HTTPException, status

from sap.beanie.exceptions import Object404Error


class Validation422Error(HTTPException):
    """Raise when querying DB returns empty result."""

    status_code: int = status.HTTP_422_UNPROCESSABLE_ENTITY
    detail: str = "The data submitted is invalid"

    def __init__(self, detail: str = ""):
        """Init exception."""
        super().__init__(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=detail or self.detail)


__all__ = [
    "Object404Error",
    "Validation422Error",
]



"""
Rest.

This package regroups all common helpers to work with a RESTful API.
"""

from .client import BeansClient, RestClient, RestData

__all__ = [
    "RestClient",
    "RestData",
    "BeansClient",
]



"""
Packet.

Packets are messages sent through a queue service to run a task on a remote server.
"""
import json
import re
import typing

import aioamqp
import aioamqp.channel
from pamqp.constants import DOMAIN_REGEX

from sap.loggers import logger
from sap.settings import SapSettings

from .amqp import AMQPClient

# from . import exceptions

DOMAIN_REGEX["queue-name"] = re.compile(r"^[a-zA-Z0-9-_.:@#,/><]*$")


class _Packet:
    """
    Define common attributes of packets.

    A Packet represents a message that has been sent to a queue
    in order to execute a task on a remote server.
    """

    namespace: str
    topic: str
    exchange: aioamqp.channel
    _is_durable: bool = True
    _exchange_type: str
    _event_type: str
    connections: dict[str, AMQPClient] = {}
    providing_args: list[str]

    def __init__(self, topic: str, providing_args: list[str]):
        """
        Initialize the packet. The topic contains the namespace.

        :providing_args: A list of arguments used for documentation purposes.
        """
        self.topic = topic
        self.namespace = topic.split(".")[0]
        self.providing_args = providing_args

    @classmethod
    async def connection_retrieve(cls) -> AMQPClient:
        """Initialize connection to AMQP and save in cache for future use."""
        if "default" not in cls.connections:
            client = AMQPClient()
            await client.connect()
            cls.connections["default"] = client
        return cls.connections["default"]

    @classmethod
    async def connection_reset(cls) -> None:
        channel = await cls.get_default_channel()
        if channel.is_open:
            await channel.close()
        cls.connections.pop("default")

    @classmethod
    async def get_default_channel(cls) -> aioamqp.channel.Channel:
        """Return the default channel for the opened connection."""
        connection: AMQPClient = await cls.connection_retrieve()
        # if not connection.channel.is_open:
        #     cls.connections.pop('default')
        #     connection = await cls.connection_retrieve()
        return connection.channel

    def exchange_get_name(self, is_fallback: bool = False) -> str:
        """
        Get exchange name.

        :is_fallback: if True, get exchange where dead packets are transferred to.
        """
        suffix: str = ".retry" if is_fallback else ""
        return f"packet.{self._event_type}{suffix}"

    async def exchange_declare(self, is_fallback: bool = False) -> None:
        """
        Declare the exchange on the AMQP server.

        :is_fallback: if True, create a fallback exchange where dead packets are transferred to.
        """
        channel = await self.get_default_channel()
        await channel.exchange_declare(
            exchange_name=self.exchange_get_name(),
            type_name=self._exchange_type,
            durable=self._is_durable,
        )
        if is_fallback:
            await channel.exchange_declare(
                exchange_name=self.exchange_get_name(is_fallback=True),
                type_name=self._exchange_type,
                durable=self._is_durable,
            )


class SignalPacket(_Packet):
    """
    A SignalPacket is a message sent to the messaging queue broker.

    to run a task asynchronously on remote server.
    Multiple applications can subscribe to the queue to receive the packets.
    """

    _is_durable: bool = True
    _exchange_type: str = "topic"
    _event_type: str = "signal"

    async def send(self, identifier: str, **kwargs: typing.Any) -> None:
        """Send the packet to the exchange."""
        if SapSettings.is_env_dev:
            logger.debug("Lambda Packet sending disabled: %s", self.topic)
            return

        assert (
            "*" not in self.topic and "#" not in self.topic
        ), "Cannot use special matching character in topic to send packet"

        await self.exchange_declare()

        channel = await self.get_default_channel()

        payload = json.dumps({"identifier": identifier, "kwargs": kwargs})

        await channel.basic_publish(
            payload=payload.encode("utf-8"),
            exchange_name=self.exchange_get_name(),
            routing_key=self.topic,
            properties={"content_type": "application/json"},
        )

    def queue_get_name(self, task_name: str, is_fallback: bool = False) -> str:
        """
        Get queue name.

        :is_fallback: if True, get queue where dead packets are transferred to.
        """
        suffix: str = "@retry" if is_fallback else ""
        return f"{self._event_type}:{self.topic}->{task_name}{suffix}"

    def queue_get_params(self, task_name: str, is_fallback: bool = False) -> dict[str, typing.Any]:
        """Retrieve params used to declare the queue."""
        name = self.queue_get_name(task_name=task_name, is_fallback=is_fallback)
        exchange_primary = self.exchange_get_name(is_fallback=False)
        exchange_fallback = self.exchange_get_name(is_fallback=True)
        arguments = {
            "x-delivery-limit": 5,
            "x-dead-letter-exchange": exchange_primary if is_fallback else exchange_fallback,
        }
        if is_fallback:
            arguments["x-message-ttl"] = 1000 * 60 * 60 * 6  # 6 hours
        else:
            arguments["x-queue-type"] = "quorum"
        return {
            "name": name,
            "exchange": exchange_fallback if is_fallback else exchange_primary,
            "routing_key": self.topic,
            "durable": True,
            "queue_arguments": arguments,
        }

    async def queue_declare(self, task_name: str) -> None:
        """Declare queues and bind them to the exchange."""
        await self.exchange_declare(is_fallback=True)

        channel = await self.get_default_channel()

        exchange_name = self.exchange_get_name()
        queue_name = self.queue_get_name(task_name)
        exchange_name_dead = self.exchange_get_name(is_fallback=True)
        queue_name_dead = self.queue_get_name(task_name, is_fallback=True)

        # A. Setup queue for dead packets
        await channel.queue_declare(
            queue_name=queue_name_dead,
            durable=True,
            arguments={
                "x-delivery-limit": 5,
                "x-message-ttl": 1000 * 60 * 60 * 6,  # 6 hours
                "x-dead-letter-exchange": exchange_name,
            },
        )
        await channel.queue_bind(queue_name=queue_name_dead, exchange_name=exchange_name_dead, routing_key=self.topic)

        # B. Setup queue for packets
        await channel.queue_declare(
            queue_name=queue_name,
            durable=True,
            arguments={
                "x-queue-type": "quorum",
                "x-delivery-limit": 5,
                "x-dead-letter-exchange": exchange_name_dead,
            },
        )
        await channel.queue_bind(queue_name=queue_name, exchange_name=exchange_name, routing_key=self.topic)



"""Settings utils."""
import os

import pydantic


class IntegrationParams(pydantic.BaseModel):
    """
    Parameters to access an external API.

    These parameters can be set directly in the environment variables.
    ex: to set the public key for Beans you can use:
    APP_SETTINGS_<CUSTOM>__BEANS_PUBLIC=tls_public_xxxxxxxxx
    """

    app_domain: str = "http://localhost:8000"
    beans_public: str = ""
    beans_secret: str = ""
    third_party_public: str = ""
    third_party_secret: str = ""
    third_party_bearer: str = ""
    is_status_available: bool = True


class DatabaseParams(pydantic.BaseModel):
    """
    Parameters to connect the database.

    These parameters can be set directly in the environment variables.
    ex: to set the host you can use for mongodb:
    APP_SETTINGS_MONGO__HOST=example.com
    """

    protocol: str = "mongodb"
    username: str = ""
    password: str = ""
    db: str = "sap"
    host: str = "localhost"
    port: str = ""
    params: str = ""

    def get_dns(self) -> str:
        """Build the DNS URI for the database."""
        credentials: str = f"{self.username}:{self.password}@" if self.username else ""
        port: str = f":{self.port}" if self.port else ""
        query: str = f"?{self.params}" if self.params else ""
        return f"{self.protocol}://{credentials}{self.host}{port}/{self.db}{query}"


class SapSettings:
    """Settings params for SAP."""

    APP_ENV: str = os.getenv("APP_ENV", "").upper()

    # True is environment is set to DEV
    is_env_dev: bool = APP_ENV == "DEV"

    # True is environment is set to PROD
    is_env_prod: bool = APP_ENV == "PROD"



"""
FastAPI.

This package regroups all common helpers exclusively
for Celery and background workers in general.
"""

from .amqp import AMQPClient
from .crons import CronStat, CronTask, FetchStrategy, register_crontask
from .lambdas import LambdaTask, LambdaWorker, register_lambda
from .packet import SignalPacket
from .utils import register_tasks_with_celery_beat

__all__ = [
    "AMQPClient",
    "SignalPacket",
    "LambdaTask",
    "LambdaWorker",
    "CronTask",
    "CronStat",
    "FetchStrategy",
    "register_lambda",
    "register_crontask",
    "register_tasks_with_celery_beat",
]



"""
Pydantic.

This package regroups all common helpers to work with Pydantic
"""
from datetime import datetime, timezone


def datetime_utcnow() -> datetime:
    """Replace deprecated datetime.utcnow."""
    return datetime.now(timezone.utc)



"""Python setup file."""
import os
from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

VERSION = "0.2.6"

setup(
    name="sapx",
    version=VERSION,
    packages=find_packages(include=("sap.*", "sap")),
    package_data={"sap": ["py.typed"]},
    include_package_data=True,
    license="COPYRIGHT @ Trellix",
    description="Library of re-usable utilities for python web apps.",
    long_description=long_description,
    url="https://github.com/trellixio/sap",
    author="Trellix Dev",
    author_email="contact@trellix.io",
    python_requires=">=3.9",
    install_requires=[
        "aioamqp~=0.15",
        "httpx~=0.26",
        "redis~=5.0",
        "celery~=5.3",
        "pydantic~=2.5",
        "PyJWT~=2.8",
        "fastAPI~=0.105",
        "itsdangerous~=2.1",
        "beanie~=1.23",
        "passlib~=1.7",
        "motor~=3.3",
        # "sqlmodel~=0.0",
        "typing-extensions~=4.9",
        "PyYAML~=6.0",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: FastAPI",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



"""
# Serializers.

Handle data validation.
"""
from __future__ import annotations

import datetime
import inspect
from typing import TYPE_CHECKING, Any, ClassVar, Generic, List, Optional, TypeVar, Union, get_args, get_origin

from typing_extensions import Literal

from fastapi import Request
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from sap.beanie.document import DocT, Document, ModelT
from sap.typing import UnionType

from .pagination import CursorInfo, PaginatedData

if TYPE_CHECKING:
    from pydantic.main import IncEx


class ObjectSerializer(BaseModel, Generic[ModelT]):
    """Serialize an object for retrieve or list."""

    @classmethod
    def get_id(cls, instance: ModelT) -> str:
        """Return the Mongo ID of the object."""
        if isinstance(instance, Document):
            return str(instance.id)

        raise NotImplementedError

    @classmethod
    def get_created(cls, instance: ModelT) -> datetime.datetime:
        """Return the user creation date."""
        if isinstance(instance, Document):
            assert instance.doc_meta.created  # let mypy know that this cannot be null
            return instance.doc_meta.created

        raise NotImplementedError

    @classmethod
    def get_updated(cls, instance: ModelT) -> datetime.datetime:
        """Return the user creation date."""
        if isinstance(instance, Document):
            assert instance.doc_meta.updated  # let mypy know that this cannot be null
            return instance.doc_meta.updated
        raise NotImplementedError

    @classmethod
    def _get_instance_data(cls, instance: ModelT, exclude: Optional["IncEx"] = None) -> dict[str, Any]:
        """Retrieve the serializer value from the instance and getters."""

        def _get_field_value(field_name: str, field_info: FieldInfo) -> Any:
            """Retrieve the value of a serialized field."""

            # A: An explicit method was declared to retrieve the value
            if hasattr(cls, f"get_{field_name}"):
                return getattr(cls, f"get_{field_name}")(instance=instance)

            related_object = getattr(instance, field_name, None)

            origin = get_origin(field_info.annotation)
            args = get_args(field_info.annotation)

            # B. The field is an embedded serializer
            if inspect.isclass(origin) and issubclass(origin, ObjectSerializer):
                return origin.read(related_object, exclude=exclude) if related_object else None

            # C. The field is a list of embedded serializer or optional
            if (
                origin in [List, Union, UnionType]
                and inspect.isclass(args[0])
                and issubclass(args[0], ObjectSerializer)
            ):
                value_default = [] if origin == List else None
                return args[0].read(related_object, exclude=exclude) if related_object else value_default

            return getattr(instance, field_name)

        data = {}
        exclude = exclude or set()
        for field_name, field_info in cls.model_fields.items():
            if field_name in exclude:
                continue
            data[field_name] = _get_field_value(field_name, field_info)
        return data

    @classmethod
    def read(cls: type["SerializerT"], instance: ModelT, exclude: Optional["IncEx"] = None) -> "SerializerT":
        """Serialize a single object instance."""
        return cls(**cls._get_instance_data(instance, exclude=exclude))

    @classmethod
    def read_list(
        cls: type["SerializerT"], instance_list: list[ModelT], exclude: Optional["IncEx"] = None
    ) -> list["SerializerT"]:
        """Serialize a list of objects."""
        return [cls.read(instance, exclude=exclude) for instance in instance_list]

    @classmethod
    def read_page(
        cls,
        instance_list: list[ModelT],
        cursor_info: CursorInfo,
        request: Request,
    ) -> PaginatedData["SerializerT"]:
        """Serialize a list of objects."""

        # TODO: implemented proper cursor pagination, for now fake it till you make it.

        page_next = cursor_info.get_next()
        page_previous = cursor_info.get_previous()
        return PaginatedData(
            count=0,
            next=str(request.url.include_query_params(cursor=page_next)) if page_next else None,
            previous=str(request.url.include_query_params(cursor=page_previous)) if page_previous else None,
            data=cls.read_list(instance_list),
        )


SerializerT = TypeVar("SerializerT", bound=ObjectSerializer[Any])


class WriteObjectSerializer(BaseModel, Generic[DocT]):
    """Serialize an object for create or update."""

    instance: Optional[DocT] = None
    embedded_serializers: ClassVar[dict[str, type["WriteObjectSerializer[Any]"]]] = {}

    def __init__(self, **data: Any) -> None:
        """Override init to filter embedded serializers."""
        super().__init__(**data)
        for field_name, field_info in self.model_fields.items():
            if inspect.isclass(field_info.annotation) and issubclass(field_info.annotation, WriteObjectSerializer):
                self.embedded_serializers[field_name] = field_info.annotation

    async def run_async_validators(self, **kwargs: Any) -> None:
        """Check that data pass DB validation."""

        # Automatically run async validators on embedded serializers
        field_serializer: WriteObjectSerializer[DocT]
        for field_name in self.embedded_serializers:
            if field_serializer := getattr(self, field_name):
                if self.instance:
                    field_serializer.instance = getattr(self.instance, field_name)
                await field_serializer.run_async_validators(**kwargs)

    def model_dump(
        self,
        *,
        mode: Literal["json", "python"] | str = "python",
        include: Optional["IncEx"] = None,
        exclude: Optional["IncEx"] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        round_trip: bool = False,
        warnings: bool = True,
    ) -> dict[str, Any]:
        """Dump the serializer data with exclusion of unwanted fields."""
        # Exclude from dumping
        if exclude:
            exclude.add("instance")  # type: ignore
        else:
            exclude = {"instance"}

        # # Some fields are only excluded from being cascade dumps to dict,
        # # but their original value is still needed
        # exclude_doc_dumps = {}

        for field_name, field_info in self.model_fields.items():
            if inspect.isclass(field_info.annotation) and issubclass(field_info.annotation, Document):
                # exclude_doc_dumps[field_name] = True
                exclude.add(field_name)  # type: ignore

        result = super().model_dump(
            mode=mode,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            round_trip=round_trip,
            warnings=warnings,
        )

        # for field_name in exclude_doc_dumps:
        #     result[field_name] = getattr(self, field_name)

        instance_embedded: Optional[BaseModel]
        for field_name, field_model in self.embedded_serializers.items():
            if not result[field_name]:
                continue
            if instance_embedded := getattr(self.instance, field_name, None):
                result[field_name] = instance_embedded.model_copy(update=result[field_name])
            else:
                result[field_name] = field_model(**result[field_name])

        return result

    async def create(self, **kwargs: Any) -> DocT:
        """Create the object in the database using the data extracted by the serializer."""
        instance_class: Optional[type[Any]] = self.model_fields["instance"].annotation
        assert instance_class and issubclass(instance_class, Document)
        self.instance = await instance_class(**self.model_dump()).create()
        return self.instance

    async def update(self, **kwargs: Any) -> DocT:
        """Update the object in the database using the data extracted by the serializer."""
        assert self.instance
        instance: DocT = self.instance.model_copy(update=self.model_dump())
        await instance.save()
        self.instance = instance
        return instance


WSerializerT = TypeVar("WSerializerT", bound=WriteObjectSerializer[Any])



"""
# Middleware.

This module contains generic middleware to perform repetitive actions on each request.
Learn more: https://fastapi.tiangolo.com/tutorial/middleware/
"""
from __future__ import annotations

import traceback
from typing import Awaitable, Callable, List, Type

import beanie
from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp, Receive, Scope, Send

from sap.beanie.client import BeanieClient
from sap.loggers import logger
from sap.settings import DatabaseParams


class LogServerErrorMiddleware(BaseHTTPMiddleware):
    """Return the server error response in a JSON object for debugging.

    Make sure that this middleware is not enable in production
    as it could leak important security information.

    There is a bug using BaseHTTPMiddleware, Jinja2, and starlette.TestClient
    https://github.com/encode/starlette/issues/472

    Using async_asgi_testclient.TestClient fix the issue while waiting for an official starlette fix
    https://github.com/tiangolo/fastapi/issues/806
    """

    async def dispatch(self, request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        """Render server error."""
        try:
            return await call_next(request)
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception(exc)
            trace = traceback.format_exception(type(exc), exc, exc.__traceback__)
            return JSONResponse(content={"traceback": trace}, status_code=500)


class InitBeanieMiddleware:
    """Middleware to initialize a connection to Mongo database."""

    app: ASGIApp
    mongo_params: DatabaseParams
    document_models: List[Type[beanie.Document]] | List[Type[beanie.View]] | List[str]
    force: bool = False

    def __init__(
        self,
        app: ASGIApp,
        mongo_params: DatabaseParams,
        document_models: List[Type[beanie.Document]] | List[Type[beanie.View]] | List[str],
        force: bool = False,
    ) -> None:
        """Initialize Middleware."""
        self.app = app
        self.mongo_params = mongo_params
        self.document_models = document_models
        self.force = force

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """Run Middleware."""
        await BeanieClient.init(
            mongo_params=self.mongo_params,
            document_models=self.document_models,
            force=self.force,
        )
        await self.app(scope, receive, send)
